package com.sun.xml.internal.bind.v2.runtime.reflect;

import java.util.concurrent.Callable;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.Patcher;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.LocatorEx;
import com.sun.istack.internal.SAXException2;
import com.sun.xml.internal.bind.WhiteSpaceProcessor;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.UnmarshallingContext;
import javax.xml.bind.JAXBException;
import javax.xml.stream.XMLStreamException;
import java.io.IOException;
import com.sun.xml.internal.bind.v2.runtime.Name;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimePropertyInfo;
import com.sun.xml.internal.bind.v2.runtime.Transducer;
import java.lang.reflect.Type;
import com.sun.xml.internal.bind.v2.runtime.reflect.opt.OptimizedTransducedAccessorFactory;
import com.sun.xml.internal.bind.v2.model.core.ID;
import com.sun.xml.internal.bind.v2.model.impl.RuntimeModelBuilder;
import com.sun.xml.internal.bind.v2.model.runtime.RuntimeNonElementRef;
import com.sun.xml.internal.bind.v2.runtime.JAXBContextImpl;
import com.sun.istack.internal.Nullable;
import com.sun.istack.internal.NotNull;
import org.xml.sax.SAXException;
import com.sun.xml.internal.bind.api.AccessorException;
import com.sun.xml.internal.bind.v2.runtime.XMLSerializer;

public abstract class TransducedAccessor<BeanT>
{
    public TransducedAccessor() {
    }
    
    public boolean useNamespace() {
        return false;
    }
    
    public void declareNamespace(BeanT o, XMLSerializer w) throws AccessorException, SAXException {
    }
    
    @Nullable
    public abstract CharSequence print(@NotNull BeanT p0) throws AccessorException, SAXException;
    
    public abstract void parse(BeanT p0, CharSequence p1) throws AccessorException, SAXException;
    
    public abstract boolean hasValue(BeanT p0) throws AccessorException;
    
    public static <T> TransducedAccessor<T> get(JAXBContextImpl context, RuntimeNonElementRef ref) {
        Transducer xducer = RuntimeModelBuilder.createTransducer(ref);
        RuntimePropertyInfo prop = ref.getSource();
        if (prop.isCollection()) {
            return new ListTransducedAccessorImpl(xducer, prop.getAccessor(), Lister.create((Type)Utils.REFLECTION_NAVIGATOR.erasure(prop.getRawType()), prop.id(), prop.getAdapter()));
        }
        if (prop.id() == ID.IDREF) {
            return new TransducedAccessor.IDREFTransducedAccessorImpl(prop.getAccessor());
        }
        if (xducer.isDefault() && context != null && !context.fastBoot) {
            TransducedAccessor<Object> xa = OptimizedTransducedAccessorFactory.get(prop);
            if (xa != null) {
                return xa;
            }
        }
        if (xducer.useNamespace()) {
            return new TransducedAccessor.CompositeContextDependentTransducedAccessorImpl(context, xducer, prop.getAccessor());
        }
        return new TransducedAccessor.CompositeTransducedAccessorImpl(context, xducer, prop.getAccessor());
    }
    
    public abstract void writeLeafElement(XMLSerializer p0, Name p1, BeanT p2, String p3) throws SAXException, AccessorException, IOException, XMLStreamException;
    
    public abstract void writeText(XMLSerializer p0, BeanT p1, String p2) throws AccessorException, SAXException, IOException, XMLStreamException;
    
    static class CompositeContextDependentTransducedAccessorImpl<BeanT, ValueT> extends TransducedAccessor.CompositeTransducedAccessorImpl<BeanT, ValueT>
    {
        public CompositeContextDependentTransducedAccessorImpl(JAXBContextImpl context, Transducer<ValueT> xducer, Accessor<BeanT, ValueT> acc) {
            super(context, xducer, acc);
            assert !(!xducer.useNamespace());
            return;
        }
        
        @Override
        public boolean useNamespace() {
            return true;
        }
        
        @Override
        public void declareNamespace(BeanT bean, XMLSerializer w) throws AccessorException {
            ValueT o = this.acc.get(bean);
            if (o != null) {
                this.xducer.declareNamespace(o, w);
            }
        }
        
        @Override
        public void writeLeafElement(XMLSerializer w, Name tagName, BeanT o, String fieldName) throws SAXException, AccessorException, IOException, XMLStreamException {
            w.startElement(tagName, null);
            this.declareNamespace(o, w);
            w.endNamespaceDecls(null);
            w.endAttributes();
            this.xducer.writeText(w, this.acc.get(o), fieldName);
            w.endElement();
        }
    }
    
    public static class CompositeTransducedAccessorImpl<BeanT, ValueT> extends TransducedAccessor<BeanT>
    {
        protected final Transducer<ValueT> xducer;
        protected final Accessor<BeanT, ValueT> acc;
        
        public CompositeTransducedAccessorImpl(JAXBContextImpl context, Transducer<ValueT> xducer, Accessor<BeanT, ValueT> acc) {
            this.xducer = xducer;
            this.acc = acc.optimize(context);
        }
        
        @Override
        public CharSequence print(BeanT bean) throws AccessorException {
            ValueT o = this.acc.get(bean);
            if (o == null) {
                return null;
            }
            return this.xducer.print(o);
        }
        
        @Override
        public void parse(BeanT bean, CharSequence lexical) throws AccessorException, SAXException {
            this.acc.set(bean, this.xducer.parse(lexical));
        }
        
        @Override
        public boolean hasValue(BeanT bean) throws AccessorException {
            return this.acc.getUnadapted(bean) != null;
        }
        
        @Override
        public void writeLeafElement(XMLSerializer w, Name tagName, BeanT o, String fieldName) throws SAXException, AccessorException, IOException, XMLStreamException {
            this.xducer.writeLeafElement(w, tagName, this.acc.get(o), fieldName);
        }
        
        @Override
        public void writeText(XMLSerializer w, BeanT o, String fieldName) throws AccessorException, SAXException, IOException, XMLStreamException {
            this.xducer.writeText(w, this.acc.get(o), fieldName);
        }
    }
    
    private static final class IDREFTransducedAccessorImpl<BeanT, TargetT> extends DefaultTransducedAccessor<BeanT>
    {
        private final Accessor<BeanT, TargetT> acc;
        private final Class<TargetT> targetType;
        
        public IDREFTransducedAccessorImpl(Accessor<BeanT, TargetT> acc) {
            this.acc = acc;
            this.targetType = acc.getValueType();
        }
        
        @Override
        public String print(BeanT bean) throws AccessorException, SAXException {
            TargetT target = this.acc.get(bean);
            if (target == null) {
                return null;
            }
            XMLSerializer w = XMLSerializer.getInstance();
            try {
                String id = w.grammar.getBeanInfo(target, true).getId(target, w);
                if (id == null) {
                    w.errorMissingId(target);
                    return id;
                }
                return id;
            }
            catch (JAXBException e) {
                w.reportError(null, e);
                return null;
            }
        }
        
        private void assign(BeanT bean, TargetT t, UnmarshallingContext context) throws AccessorException {
            if (!this.targetType.isInstance(t)) {
                context.handleError(Messages.UNASSIGNABLE_TYPE.format(new Object[] { this.targetType, t.getClass() }));
            }
            else {
                this.acc.set(bean, t);
            }
        }
        
        @Override
        public void parse(BeanT bean, CharSequence lexical) throws AccessorException, SAXException {
            String idref = WhiteSpaceProcessor.trim(lexical).toString();
            UnmarshallingContext context = UnmarshallingContext.getInstance();
            Callable<Object> callable = context.getObjectFromId(idref, this.acc.valueType);
            if (callable == null) {
                context.errorUnresolvedIDREF(bean, idref, context.getLocator());
                return;
            }
            TargetT t;
            try {
                t = callable.call();
            }
            catch (SAXException e) {
                throw e;
            }
            catch (RuntimeException e2) {
                throw e2;
            }
            catch (Exception e3) {
                throw new SAXException2(e3);
            }
            if (t != null) {
                this.assign(bean, t, context);
            }
            else {
                LocatorEx loc = new LocatorEx.Snapshot(context.getLocator());
                context.addPatcher(new Patcher() {
                    public void run() throws SAXException {
                        try {
                            TargetT t = callable.call();
                            if (t == null) {
                                context.errorUnresolvedIDREF(bean, idref, loc);
                            }
                            else {
                                TransducedAccessor.IDREFTransducedAccessorImpl.access$000(TransducedAccessor.IDREFTransducedAccessorImpl.this, bean, t, context);
                            }
                        }
                        catch (AccessorException e) {
                            context.handleError(e);
                        }
                        catch (SAXException e2) {
                            throw e2;
                        }
                        catch (RuntimeException e3) {
                            throw e3;
                        }
                        catch (Exception e4) {
                            throw new SAXException2(e4);
                        }
                    }
                });
            }
            return;
        }
        
        @Override
        public boolean hasValue(BeanT bean) throws AccessorException {
            return this.acc.get(bean) != null;
        }
        
        static void access$000(TransducedAccessor.IDREFTransducedAccessorImpl x0, Object x1, Object x2, UnmarshallingContext x3) throws AccessorException {
            x0.assign(x1, x2, x3);
        }
    }
}
